#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : detect.py
# Contents: detect synchronous spiking activity and reduce results
# Author  : Christian Borgelt
# History : 2013.02.13 file created from detect/detect.py
#           2013.02.23 added jitter to injected coincidences
#           2013.02.25 extended to larger grid 2,...,12 x 2,...,12
#           2013.02.27 combined with script for false negatives
#           2013.04.29 jitter added to output file names
#           2013.05.07 added false negative rates for raw patterns
#           2013.05.08 adapted some file names, removed filtered result
#           2013.12.06 adapted to filtering with detection border
#           2014.05.27 modules placed in sibling directories
#           2015.08.12 adapted to modified pattern set reduction
#           2015.08.18 changed to function patred() from pycoco library
#-----------------------------------------------------------------------
from sys     import argv, path
from os      import makedirs
from os.path import join
from time    import strftime
from random  import seed
UTILDIR  = join('..', 'util')
if UTILDIR  not in path: path.insert(0, UTILDIR)
from id2par  import get_args, project, run
NEURODIR = join('..', 'neuro')
if NEURODIR not in path: path.insert(0, NEURODIR)
from psts    import genpsts
try:
    from coco    import coconad, patred
except ImportError:
    COCODIR = join('..', 'coconad')
    if COCODIR not in path: path.insert(0, COCODIR)
    from coconad import coconad
    FIMDIR  = join('..', 'fim')
    if FIMDIR   not in path: path.insert(0, FIMDIR)
    from patred  import patred
    print('Attention: Could not import from PyCoCo library!')
    print('           Using Python replacement (very slow)!')

#-----------------------------------------------------------------------

oo = float('inf')               # positive infinity as a constant
redids = ['x', 'c', 'C', 'i', 's', 'S', 'l', 'L', 't', 'T']

#-----------------------------------------------------------------------

def detect (ncnt=100, period=3, rate=20,         # neuron    parameters
            size=5, coin=5, jitter=0.003,        # assembly  parameters
            width=0.003, target='c', rfnids=[0], # detection parameters
            runs=1000, verbose=100, srng=1):     # other     parameters
    if jitter < 0: jitter = width
    srng += 1023*size           # init. seed and get data file name
    base = 'n%d_p%g_r%g_w%g_%c' % (ncnt,period,rate,width*1000,target)
    psp  = set()                # initialize a pattern spectrum
    with open(join('data','pc_'+base+'.txt'), 'r') as inp:
        for line in inp:        # collect the pattern signatures
            z,c,n = line.split()
            psp.add((int(z),int(c)))
    border = list(range(ncnt+1)); m = 1
    for z in reversed(border):  # determine the detection border
        m = max([c for a,c in psp if a == z]+[m])
        border[z] = m+1         # (support thresholds per size)
    border[0:2] = [oo,oo]       # entirely rule out sizes 0 and 1
    base = 'n%d_p%g_r%g_j%g_w%g_%c' \
         % (ncnt,period,rate,jitter*1000,width*1000,target)
    ext  = '_%d_%d' % (size,coin)
    name = base +ext            # form the file name (parts)
    asm  = set(range(1,size+1)) # create the assembly to inject
    upcs = dict()               # and the pattern count dictionaries
    fpcs = dict()               # (unfiltered, filtered, reduced)
    rpcs = dict([(r,dict()) for r in rfnids])
    draw = xraw = 0                      # runs with raw detections
    dect = dict([(r,0) for r in rfnids]) # runs with detections
    xdct = dict([(r,0) for r in rfnids]) # runs with exact detections
    rexs = dict([(r,0) for r in rfnids]) # runs with excess neurons
    rmis = dict([(r,0) for r in rfnids]) # runs with no detect patterns
    rops = dict([(r,0) for r in rfnids]) # runs with all other patterns
    sups = dict([(r,0) for r in rfnids]) # number of superset  patterns
    subs = dict([(r,0) for r in rfnids]) # number of subset    patterns
    ovls = dict([(r,0) for r in rfnids]) # number of overlap   patterns
    urls = dict([(r,0) for r in rfnids]) # number of unrelated patterns
    for i in range(runs):       # repeat the experiment multiple times
        if verbose > 0 and i % verbose == 0:
            print(strftime('%Y.%m.%d %H:%M:%S detect ')+name+' '+str(i))
        seed(srng)              # seed random number generator
        psts = genpsts(ncnt, period, rate, jitter=jitter,
                       asms=[[asm,-coin]])
        pats = coconad(list(zip(range(1,ncnt+1), psts)),
                       target=target, width=width, supp=2, zmin=2)
        fltd = []; d = e = 0    # generate and mine spike trains
        for p,c in pats:        # and traverse the found patterns
            z = len(p)          # get the pattern size
            p = frozenset(p)    # and the set of items/neurons
            if p >= asm:        # check for raw detections
                if z > size: e = 1     # superset detection
                else:        d = 1     # exact    detection
            s = (z,c)           # create the pattern signature
            if s in upcs: upcs[s] += 1 # count unfiltered patterns
            else:         upcs[s]  = 1 # by their signatures
            if c < border[z]: continue
            if s in fpcs: fpcs[s] += 1 # count filtered patterns
            else:         fpcs[s]  = 1 # by their signatures
            fltd.append((p,c))  # collect non-excluded patterns
        draw += d | e           # count runs with raw detections
        xraw += d               # (superset and exact)
        for r in rfnids:        # traverse the reduction function ids
            pats  = patred(fltd, redids[r], border, True)
            pcs   = rpcs[r]     # reduce the filtered patterns
            d = e = x = y = 0   # initialize the detection flags
            for p,c in pats:    # traverse the reduced patterns
                z = len(p)      # get the pattern size and
                s = (z,c)       # create the pattern signature
                if s in pcs: pcs[s] += 1 # count reduced patterns
                else:        pcs[s]  = 1 # by their signatures
                if   p >= asm:  # set detection flags, count patterns
                    if z > size:       sups[r] += 1; e = 1
                    else:                            d = 1
                elif p <  asm:         subs[r] += 1; x = 1
                elif len(p & asm) > 1: ovls[r] += 1; x = 1
                else:                  urls[r] += 1; x = 1
            dect[r] += d | e    # count runs with detections,
            xdct[r] += d        # runs with exact detections,
            rexs[r] += e        # runs with excess neurons,
            rmis[r] += x        # runs with no detect patterns,
            rops[r] += x | e    # runs with all other patterns
        srng += coin            # switch to the next random seed
    if verbose > 0:             # print termination log message
        print(strftime('%Y.%m.%d %H:%M:%S detect ')
                      +name+' '+str(runs)+' done.')
    try: makedirs('data')       # ensure that a directory exists
    except OSError: pass        # for the result files
    n = 1.0/runs if runs > 0 else 1.0
    with open(join('data','pu_'+name+'.txt'), 'w') as out:
        for x in sorted([(s[0], s[1], upcs[s]*n) for s in upcs]):
            out.write('%g %g %g\n' % x)
    with open(join('data','pf_'+name+'.txt'), 'w') as out:
        for x in sorted([(s[0], s[1], fpcs[s]*n) for s in fpcs]):
            out.write('%g %g %g\n' % x)
    with open(join('data','fn_'+name+'.txt'), 'w') as out:
        out.write('%g '  % (1.0 -draw*n))
        out.write('%g\n' % (1.0 -xraw*n))
    for r in rfnids:            # traverse the reduction function ids
        pcs = rpcs[r]           # and write the pattern counters
        with open(join('data','pr_'+base+str(r)+ext+'.txt'),'w') as out:
            for x in sorted([(s[0], s[1], pcs[s]*n) for s in pcs]):
                out.write('%g %g %g\n' % x)
        with open(join('data','fn_'+base+str(r)+ext+'.txt'),'w') as out:
            out.write('%g '  % (1.0 -dect[r]*n))
            out.write('%g '  % (1.0 -xdct[r]*n))
            out.write('%g '  % (rexs[r]*n)) # write the average
            out.write('%g '  % (rmis[r]*n)) # pattern counts
            out.write('%g '  % (rops[r]*n))
            out.write('%g '  % (sups[r]*n))
            out.write('%g '  % (subs[r]*n))
            out.write('%g '  % (ovls[r]*n))
            out.write('%g\n' % (urls[r]*n))
    return (upcs, fpcs, rpcs,   # return the pattern spectrums
            dect, xdct, rexs, rmis, rops, sups, subs, ovls, urls)

#-----------------------------------------------------------------------

def collect (ncnt=100, period=3, rate=20,          # neuron    params.
             sizes=[5], coins=[5], jitter=0.003,   # assembly  params.
             width=0.003, target='c', rfnids=[4]): # detection params.
    if jitter < 0: jitter = width
    base = 'n%d_p%g_r%g_j%g_w%g_%c' \
         % (ncnt,period,rate,jitter*1000,width*1000,target)
    fneg = dict()               # get the pattern count file name base
    fnxs = dict()               # and initialize run count dictionaries
    for s in [(z,c) for z in sizes for c in coins]:
        ext = '_%d_%d' % s      # read raw false negatives
        with open(join('data','fn_'+base+ext+'.txt'), 'r') as inp:
            fneg[s],fnxs[s] = inp.read()[:-1].split()
    with open(join('data','fn_'+base+'.txt'), 'w') as out:
        for x in sorted([(k[0], k[1], fneg[k]) for k in fneg]):
            out.write('%g %g %s\n' % x)
    with open(join('data','fx_'+base+'.txt'), 'w') as out:
        for x in sorted([(k[0], k[1], fnxs[k]) for k in fnxs]):
            out.write('%g %g %s\n' % x)
    for r in rfnids:            # traverse the reduction function ids
        name = base +str(r)     # get the pattern count file name
        fneg = dict(); fnxs = dict()
        rexs = dict(); rmis = dict(); rops = dict()
        sups = dict(); subs = dict(); ovls = dict(); urls = dict()
        for s in [(z,c) for z in sizes for c in coins]:
            ext = '_%d_%d' % s  # read raw false negatives
            with open(join('data','fn_'+name+ext+'.txt'), 'r') as inp:
                vals = inp.read()[:-1].split()
                fneg[s],fnxs[s]                 = vals[0:2]
                rexs[s],rmis[s],rops[s]         = vals[2:5]
                sups[s],subs[s],ovls[s],urls[s] = vals[5:9]
        with open(join('data','fn_'+name+'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], fneg[k]) for k in fneg]):
                out.write('%g %g %s\n' % x)
        with open(join('data','fx_'+name+'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], fnxs[k]) for k in fnxs]):
                out.write('%g %g %s\n' % x)
        with open(join('data','rx_'+name+'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], rexs[k]) for k in rexs]):
                out.write('%g %g %s\n' % x)
        with open(join('data','rn_'+name+'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], rmis[k]) for k in rmis]):
                out.write('%g %g %s\n' % x)
        with open(join('data','ro_'+name+'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], rops[k]) for k in rops]):
                out.write('%g %g %s\n' % x)
        with open(join('data','sp_'+name+'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], sups[k]) for k in sups]):
                out.write('%g %g %s\n' % x)
        with open(join('data','sb_'+name+'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], subs[k]) for k in subs]):
                out.write('%g %g %s\n' % x)
        with open(join('data','ov_'+name+'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], ovls[k]) for k in ovls]):
                out.write('%g %g %s\n' % x)
        with open(join('data','ur_'+name+'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], urls[k]) for k in urls]):
                out.write('%g %g %s\n' % x)

#-----------------------------------------------------------------------

if __name__ == '__main__':
    grids = [ [ ( 'ncnt',   [  100 ] ),
                ( 'period', [    3 ] ),
                ( 'rate',   [   20 ] ),
                ( 'size',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ] ),
                ( 'coin',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ] ),
                ( 'target', [  'c' ] ),
                ( {'jitter':0.005, 'width':0.005},
                  {'jitter':0.004, 'width':0.004},
                  {'jitter':0.003, 'width':0.003},
                  {'jitter':0.002, 'width':0.002} ),
                ( 'rfnids', [ range(len(redids)) ] ),
                ( 'runs',   [ 1000 ] ) ] ]

    if len(argv) < 2 or argv[1] != 'collect':
        run(detect, grids, int(argv[1])-1 if len(argv) > 1 else -1)
    else:
        ranges = dict(project(grids, ['size','coin'])[0])
        grids  = project(grids, ['ncnt','period','rate','width',
                                 'jitter','target','rfnids'])
        for i in range(get_args(grids)):
            collect(sizes=ranges['size'], coins=ranges['coin'],
                    **get_args(grids, i))
